# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class PatchTestCase(MididingsTestCase):

    def test_unit_order(self):
        def me_impl(ev, n):
            order.append(n)
            return ev

        def me(n):
            return Process(lambda ev: me_impl(ev, n))

        order = []
        p = ([ me(1), me(2) ] >>
            me(3) >> [
                me(4),
                me(5) >> [ me(6), me(7) ] >> me(8)
            ] >> me(9))
        self.run_patch(p, self.make_event())
        self.assertEqual(order, [1, 2, 3, 4, 5, 6, 7, 8, 9])

        def fork_dup(*args):
            return Fork(*args, remove_duplicates=False)

        order = []
        p = (fork_dup([ me(1), me(2) ]) >>
            me(3) >> fork_dup([
                me(4),
                me(5) >> fork_dup([ me(6), me(7) ]) >> me(8)
            ]) >> me(9))

        self.run_patch(p, self.make_event())
        self.assertEqual(order, [1, 2, 3, 3, 4, 5, 6, 7, 8, 8,
                                 4, 5, 6, 7, 8, 8, 9, 9, 9, 9, 9, 9])



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *

import threading


class CallTestCase(MididingsTestCase):

    @data_offsets
    def test_Process(self, off):
        def foo(ev):
            self.assertEqual(ev.type, NOTEON)
            self.assertEqual(ev.port, off(0))
            self.assertEqual(ev.channel, off(0))
            self.assertEqual(ev.note, 66)
            self.assertEqual(ev.velocity, 23)
            ev.type = CTRL
            ev.port = off(4)
            ev.channel = off(5)
            ev.ctrl = 23
            ev.value = 42
            return ev

        self.check_patch(Process(foo), {
            self.make_event(NOTEON, off(0), off(0), 66, 23):
                [self.make_event(CTRL, off(4), off(5), 23, 42)],
        })

    def test_Process_return(self):
        ev = self.make_event()

        self.check_patch(Process(lambda ev: ev), {
            ev: [ev]
        })

        self.check_patch(Process(lambda ev: None), {
            ev: []
        })

        self.check_patch(Process(lambda ev: []), {
            ev: []
        })

        self.check_patch(Process(lambda ev: [ev, ev, ev]), {
            ev: [ev, ev, ev]
        })

        self.check_patch(Process(lambda ev: (ev for ev in [ev, ev])), {
            ev: [ev, ev]
        })

    @data_offsets
    def test_Call(self, off):
        event = threading.Event()

        def foo(ev):
            self.assertEqual(ev.type, NOTEON)
            self.assertEqual(ev.port, off(0))
            self.assertEqual(ev.channel, off(0))
            self.assertEqual(ev.note, 66)
            self.assertEqual(ev.velocity, 23)
            event.set()

        ev = self.make_event(NOTEON, off(0), off(0), 66, 23)
        self.check_patch(Call(foo), { ev: [] })
        self.assertTrue(event.wait(1.0))

    @data_offsets
    def test_Call_no_arg(self, off):
        event = threading.Event()

        def foo():
            event.set()

        ev = self.make_event()
        self.check_patch(Call(foo), { ev: [] })
        self.assertTrue(event.wait(1.0))

    @data_offsets
    def test_Call_partial(self, off):
        event = threading.Event()

        def foo(ev, bar, **kwargs):
            self.assertEqual(ev.type, NOTEON)
            self.assertEqual(ev.port, off(0))
            self.assertEqual(ev.channel, off(0))
            self.assertEqual(ev.note, 66)
            self.assertEqual(ev.velocity, 23)
            self.assertEqual(bar, 42)
            self.assertEqual(kwargs['baz'], 666)
            event.set()

        ev = self.make_event(NOTEON, off(0), off(0), 66, 23)
        self.check_patch(Call(foo, 42, baz=666), { ev: [] })
        self.assertTrue(event.wait(1.0))

    def test_Call_method(self):
        event = threading.Event()

        class Foo(object):
            def bar(myself, ev):
                self.assertEqual(ev.type, NOTEON)
                event.set()
        obj = Foo()

        ev = self.make_event(NOTEON)
        self.check_patch(Call(obj.bar), { ev: [] })
        self.assertTrue(event.wait(1.0))

    def test_Call_callable(self):
        event = threading.Event()

        class Foo(object):
            def __call__(myself, ev):
                self.assertEqual(ev.type, NOTEON)
                event.set()
        obj = Foo()

        ev = self.make_event(NOTEON)
        self.check_patch(Call(obj), { ev: [] })
        self.assertTrue(event.wait(1.0))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *

import random


class ModifiersTestCase(MididingsTestCase):

    @data_offsets
    def test_Port(self, off):
        ev = self.make_event()

        self.check_patch(Port(off(7)), {
            ev: [self.modify_event(ev, port=off(7))],
        })
        self.check_patch(Port(port=off(8)), {
            ev: [self.modify_event(ev, port=off(8))],
        })

        with self.assertRaises(TypeError):
            Port(123.456)

    @data_offsets
    def test_Channel(self, off):
        ev = self.make_event()

        self.check_patch(Channel(off(7)), {
            ev: [self.modify_event(ev, channel=off(7))],
        })
        self.check_patch(Channel(channel=off(8)), {
            ev: [self.modify_event(ev, channel=off(8))],
        })

        with self.assertRaises(TypeError):
            Channel(123.456)
        with self.assertRaises(TypeError):
            Channel('blah')
        with self.assertRaises(ValueError):
            Channel(17)

    @data_offsets
    def test_Transpose(self, off):
        ev1 = self.make_event(NOTEON, note=random.randrange(0, 115))
        ev2 = self.make_event(NOTEOFF, note=random.randrange(0, 115))
        ev3 = self.make_event(NOTEON, note=random.randrange(24, 127))
        ev4 = self.make_event(NOTEOFF, note=random.randrange(24, 127))
        ev5 = self.make_event(CTRL)
        ev6 = self.make_event(POLY_AFTERTOUCH, note=random.randrange(0, 115))
        ev7 = self.make_event(POLY_AFTERTOUCH, note=random.randrange(24, 127))

        self.check_patch(Transpose(12), {
            ev1: [self.modify_event(ev1, note=ev1.note + 12)],
            ev2: [self.modify_event(ev2, note=ev2.note + 12)],
            ev5: [ev5],
            ev6: [self.modify_event(ev6, note=ev6.note + 12)],
        })

        self.check_patch(Transpose(octaves=-2), {
            ev3: [self.modify_event(ev3, note=ev3.note - 24)],
            ev4: [self.modify_event(ev4, note=ev4.note - 24)],
            ev5: [ev5],
            ev7: [self.modify_event(ev7, note=ev7.note - 24)],
        })

        with self.assertRaises(TypeError):
            Transpose(123.456)
        with self.assertRaises(TypeError):
            Transpose('blah')

    @data_offsets
    def test_Key(self, off):
        ev = self.make_event(NOTE)
        ev2 = self.make_event(CTRL)
        ev3 = self.make_event(POLY_AFTERTOUCH)

        self.check_patch(Key(42), {
            ev: [self.modify_event(ev, note=42)],
            ev2: [ev2],
            ev3: [self.modify_event(ev3, note=42)],
        })

        with self.assertRaises(TypeError):
            Key(123.456)
        with self.assertRaises(ValueError):
            Key('blah')
        with self.assertRaises(ValueError):
            Key(-1)
        with self.assertRaises(ValueError):
            Key(128)

    @data_offsets
    def test_Velocity(self, off):
        ev = self.make_event(NOTEON, velocity=42)
        ev2 = self.make_event(PROGRAM)

        self.check_patch(Velocity(23), {
            ev: [self.modify_event(ev, velocity=65)],
            ev2: [ev2],
        })

        self.check_patch(Velocity(multiply=1.5), {
            ev: [self.modify_event(ev, velocity=63)],
            ev2: [ev2],
        })

        self.check_patch(Velocity(fixed=108), {
            ev: [self.modify_event(ev, velocity=108)],
            ev2: [ev2],
        })

        self.check_patch(Velocity(gamma=2.0), {
            ev: [self.modify_event(ev, velocity=73)],
            ev2: [ev2],
        })

        self.check_patch(Velocity(curve=-2.0), {
            ev: [self.modify_event(ev, velocity=18)],
            ev2: [ev2],
        })

        self.check_patch(Velocity(multiply=1.5, offset=23), {
            ev: [self.modify_event(ev, velocity=86)],
            ev2: [ev2],
        })

        with self.assertRaises(TypeError):
            Velocity('blah')
        with self.assertRaises(TypeError):
            Velocity(something=123)

    @data_offsets
    def test_VelocitySlope(self, off):
        ev0 = self.make_event(NOTEON, note=0, velocity=42)
        ev23 = self.make_event(NOTEON, note=23, velocity=42)
        ev32 = self.make_event(NOTEON, note=32, velocity=42)
        ev42 = self.make_event(NOTEON, note=42, velocity=42)
        ev127 = self.make_event(NOTEON, note=127, velocity=42)

        self.check_patch(VelocitySlope((23, 42), (-13, +13)), {
            ev0: [self.modify_event(ev0, velocity=29)],
            ev23: [self.modify_event(ev23, velocity=29)],
            ev32: [self.modify_event(ev32, velocity=42)],
            ev42: [self.modify_event(ev42, velocity=55)],
            ev127: [self.modify_event(ev127, velocity=55)],
        })

        self.check_patch(VelocitySlope(('b-1', 'f#1'), multiply=(0.5, 2.0)), {
            ev0: [self.modify_event(ev0, velocity=21)],
            ev23: [self.modify_event(ev23, velocity=21)],
            ev32: [self.modify_event(ev32, velocity=50)],
            ev42: [self.modify_event(ev42, velocity=84)],
            ev127: [self.modify_event(ev127, velocity=84)],
        })

        self.check_patch(VelocitySlope((23, 42), fixed=(13, 113)), {
            ev0: [self.modify_event(ev0, velocity=13)],
            ev23: [self.modify_event(ev23, velocity=13)],
            ev32: [self.modify_event(ev32, velocity=60)],
            ev42: [self.modify_event(ev42, velocity=113)],
            ev127: [self.modify_event(ev127, velocity=113)],
        })

        self.check_patch(VelocitySlope(('b-1', 'f#1'), gamma=(1.0, 2.0)), {
            ev0: [self.modify_event(ev0, velocity=42)],
            ev23: [self.modify_event(ev23, velocity=42)],
            ev32: [self.modify_event(ev32, velocity=60)],
            ev42: [self.modify_event(ev42, velocity=73)],
            ev127: [self.modify_event(ev127, velocity=73)],
        })

        self.check_patch(VelocitySlope((23, 42), curve=(0, -2.0)), {
            ev0: [self.modify_event(ev0, velocity=42)],
            ev23: [self.modify_event(ev23, velocity=42)],
            ev32: [self.modify_event(ev32, velocity=29)],
            ev42: [self.modify_event(ev42, velocity=18)],
            ev127: [self.modify_event(ev127, velocity=18)],
        })

        self.check_patch(VelocitySlope((23, 42), (0.5, 2.0), (-13, +13)), {
            ev0: [self.modify_event(ev0, velocity=8)],
            ev23: [self.modify_event(ev23, velocity=8)],
            ev32: [self.modify_event(ev32, velocity=50)],
            ev42: [self.modify_event(ev42, velocity=97)],
            ev127: [self.modify_event(ev127, velocity=97)],
        })

        with self.assertRaises(TypeError):
            VelocitySlope((23, 42, 66.6), (4, 8, 15))
        with self.assertRaises(ValueError):
            VelocitySlope((23, 42, 66), (4, 8))
        with self.assertRaises(ValueError):
            VelocitySlope((23, 66, 42), (4, 8, 15))

    @data_offsets
    def test_VelocityLimit(self, off):
        ev23 = self.make_event(NOTEON, velocity=23)
        ev42 = self.make_event(NOTEON, velocity=42)
        ev108 = self.make_event(NOTEON, velocity=108)
        ev_noteoff = self.make_event(NOTEOFF)
        ev_ctrl = self.make_event(CTRL)

        self.check_patch(VelocityLimit(32, 69), {
            ev23: [self.modify_event(ev23, velocity=32)],
            ev42: [self.modify_event(ev42, velocity=42)],
            ev108: [self.modify_event(ev108, velocity=69)],
            ev_noteoff: [ev_noteoff],
            ev_ctrl: [ev_ctrl],
        })

        self.check_patch(VelocityLimit(max=69), {
            ev23: [self.modify_event(ev23, velocity=23)],
            ev42: [self.modify_event(ev42, velocity=42)],
            ev108: [self.modify_event(ev108, velocity=69)],
            ev_noteoff: [ev_noteoff],
            ev_ctrl: [ev_ctrl],
        })

        self.check_patch(VelocityLimit(min=32), {
            ev23: [self.modify_event(ev23, velocity=32)],
            ev42: [self.modify_event(ev42, velocity=42)],
            ev108: [self.modify_event(ev108, velocity=108)],
            ev_noteoff: [ev_noteoff],
            ev_ctrl: [ev_ctrl],
        })

    @data_offsets
    def test_CtrlMap(self, off):
        ev1 = self.make_event(CTRL, ctrl=23)
        ev2 = self.make_event(CTRL, ctrl=42)
        ev3 = self.make_event(PROGRAM)

        self.check_patch(CtrlMap(23, 69), {
            ev1: [self.modify_event(ev1, ctrl=69)],
            ev2: [ev2],
            ev3: [ev3],
        })

    @data_offsets
    def test_CtrlRange(self, off):
        ev1 = self.make_event(CTRL, ctrl=23, value=0)
        ev2 = self.make_event(CTRL, ctrl=23, value=64)
        ev3 = self.make_event(CTRL, ctrl=23, value=127)

        self.check_patch(CtrlRange(23, 42, 69), {
            ev1: [self.modify_event(ev1, value=42)],
            ev2: [self.modify_event(ev2, value=55)],
            ev3: [self.modify_event(ev3, value=69)],
        })

        self.check_patch(CtrlRange(23, 42, 69, 64, 127), {
            ev1: [self.modify_event(ev1, value=42)],
            ev2: [self.modify_event(ev2, value=42)],
            ev3: [self.modify_event(ev3, value=69)],
        })

    @data_offsets
    def test_CtrlCurve(self, off):
        ev = self.make_event(CTRL, ctrl=23, value=42)
        ev2 = self.make_event(AFTERTOUCH)

        self.check_patch(CtrlCurve(23, offset=23), {
            ev: [self.modify_event(ev, value=65)],
            ev2: [ev2],
        })

        self.check_patch(CtrlCurve(23, multiply=1.5), {
            ev: [self.modify_event(ev, value=63)],
            ev2: [ev2],
        })

        self.check_patch(CtrlCurve(23, gamma=2.0), {
            ev: [self.modify_event(ev, value=73)],
            ev2: [ev2],
        })

        self.check_patch(CtrlCurve(23, curve=-2.0), {
            ev: [self.modify_event(ev, value=18)],
            ev2: [ev2],
        })

        self.check_patch(CtrlCurve(23, 1.5, 23), {
            ev: [self.modify_event(ev, value=86)],
            ev2: [ev2],
        })

    @data_offsets
    def test_Pitchbend(self, off):
        ev1 = self.make_event(PITCHBEND, value=-8192)
        ev2 = self.make_event(PITCHBEND, value=0)
        ev3 = self.make_event(PITCHBEND, value=8191)

        self.check_patch(PitchbendRange(-1234, 5678), {
            ev1: [self.modify_event(ev1, value=-1234)],
            ev2: [self.modify_event(ev2, value=0)],
            ev3: [self.modify_event(ev3, value=5678)],
        })

        self.check_patch(PitchbendRange(-12, 2, range=12), {
            ev1: [self.modify_event(ev1, value=-8192)],
            ev2: [self.modify_event(ev2, value=0)],
            ev3: [self.modify_event(ev3, value=1365)],
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class SplitsTestCase(MididingsTestCase):

    @data_offsets
    def test_PortSplit(self, off):
        p = PortSplit({
            off(0): Discard(),
            (off(1), off(2)): Pass()
        })
        self.check_patch(p, {
            self.make_event(port=off(0)): False,
            self.make_event(port=off(1)): True,
        })

        p = PortSplit({
            off(0): Discard(),
            (off(2), off(3)): Discard(),
            None: Pass()
        })
        self.check_patch(p, {
            self.make_event(port=off(0)): False,
            self.make_event(port=off(1)): True,
        })

    @data_offsets
    def test_ChannelSplit(self, off):
        p = ChannelSplit({
            off(0): Discard(),
            (off(1), off(2)): Pass()
        })
        self.check_patch(p, {
            self.make_event(channel=off(0)): False,
            self.make_event(channel=off(1)): True,
        })

        p = ChannelSplit({
            off(0): Discard(),
            (off(2), off(3)): Discard(),
            None: Pass()
        })
        self.check_patch(p, {
            self.make_event(channel=off(0)): False,
            self.make_event(channel=off(1)): True,
        })

    @data_offsets
    def test_KeySplit(self, off):
        ev1 = self.make_event(NOTEON, note=66)
        ev2 = self.make_event(NOTEON, note=42)
        ev3 = self.make_event(PROGRAM)

        p = KeySplit(55, Channel(off(3)), Channel(off(7)))
        self.check_patch(p, {
            ev1: [self.modify_event(ev1, channel=off(7))],
            ev2: [self.modify_event(ev2, channel=off(3))],
            ev3: [self.modify_event(ev3, channel=off(3)),
                  self.modify_event(ev3, channel=off(7))],
        })

        p = KeySplit({
            (23, 52): Pass(),
            (52, 69): Discard(),
            (69, 88): Pass(),
        })
        self.check_patch(p, {
            ev1: False,
            ev2: True,
            ev3: True,
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

try:
    import unittest2 as unittest
except:
    import unittest

from mididings import arguments
from mididings import misc


class ArgumentsTestCase(unittest.TestCase):

    def test_simple(self):
        @arguments.accept(int)
        def foo(a): pass

        foo(123)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123.456)

        @arguments.accept(str, float)
        def bar(a, b): pass

        bar('blah', 123.456)

        with self.assertRaises(TypeError):
            bar()
        with self.assertRaises(TypeError):
            bar('blah')
        with self.assertRaises(TypeError):
            bar('blah', 123)

    def test_isinstance(self):
        @arguments.accept((int, str))
        def foo(a): pass

        foo(123)
        foo('blah')

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123.456)

    def test_value(self):
        @arguments.accept((True, 42))
        def foo(a): pass

        foo(True)
        foo(42)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(ValueError):
            foo(False)
        with self.assertRaises(ValueError):
            foo(23)
        with self.assertRaises(ValueError):
            foo('blah')

    def test_varargs(self):
        @arguments.accept(int, int)
        def foo(a, *args): pass

        foo(123)
        foo(123, 456, 789)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123, 456, 'blah')

    def test_kwargs(self):
        @arguments.accept(kwargs={'a': int, 'b': str})
        def foo(**kwargs): pass

        foo()
        foo(a=123)
        foo(b='blah')
        foo(a=123, b='blah')

        with self.assertRaises(TypeError):
            foo(a='blah')
        with self.assertRaises(TypeError):
            foo(b=123)
        with self.assertRaises(TypeError):
            foo(c=123)

        @arguments.accept(int, kwargs={'b': str, 'c': float})
        def bar(a, **kwargs):
            self.assertEqual(a, 123)
            self.assertDictEqual(kwargs, {'b': 'blah', 'c': 123.456})

        bar(123, b='blah', c=123.456)
        bar(a=123, b='blah', c=123.456)

        with self.assertRaises(TypeError):
            bar(123, b=456, c=789)
        with self.assertRaises(TypeError):
            bar(123, b=456, c=789, d=123)

    def test_kwargs_any(self):
        @arguments.accept(None, kwargs={None: int})
        def foo(*args, **kwargs):
            self.assertEqual(len(args), 2)
            self.assertEqual(args[0], 23)
            self.assertEqual(args[1], 42)
            self.assertEqual(len(kwargs), 2)
            self.assertEqual(kwargs['bar'], 13)
            self.assertEqual(kwargs['baz'], 666)

        foo(23, 42, bar=13, baz=666)

        with self.assertRaises(TypeError):
            foo(23, 42, bar=13, baz=66.6)

    def test_nullable(self):
        @arguments.accept(arguments.nullable(int))
        def foo(a): pass

        foo(None)
        foo(42)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123.456)

    def test_sequenceof(self):
        @arguments.accept(arguments.sequenceof(int))
        def foo(a):
            self.assertTrue(misc.issequenceof(a, int))

        foo([])
        foo([123, 456, 789])

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123, 456, 789)
        with self.assertRaises(TypeError):
            foo([123, 456, 'blah'])

        @arguments.accept([str])
        def bar(a):
            self.assertTrue(misc.issequenceof(a, str))

        bar([])
        bar(['doo', 'bee', 'doo'])

        with self.assertRaises(TypeError):
            bar()
        with self.assertRaises(TypeError):
            bar('doo', 'bee', 'doo')
        with self.assertRaises(TypeError):
            bar('doo', 'bee', 123)

    def test_tupleof(self):
        @arguments.accept(arguments.tupleof(int, str))
        def foo(a): pass

        foo((123, 'blah'))
        foo([123, 'blah'])

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123, 'blah')
        with self.assertRaises(TypeError):
            foo((123, 456))
        with self.assertRaises(ValueError):
            foo([123, 'blah', 789])

        @arguments.accept([int, float])
        def bar(a): pass

        bar((123, 456.789))
        bar([123, 456.789])

        with self.assertRaises(TypeError):
            bar()
        with self.assertRaises(TypeError):
            bar(123, 456.789)
        with self.assertRaises(TypeError):
            bar((123, 'blah'))
        with self.assertRaises(ValueError):
            bar([123, 456.789, 'blah'])

    def test_mappingof(self):
        @arguments.accept(arguments.mappingof(int, float))
        def foo(a):
            for k, v in a.items():
                self.assertIsInstance(k, int)
                self.assertIsInstance(v, float)

        foo({})
        foo({23: 12.34, 42: 56.78})

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo({23: 12.34, 42: 666})
        with self.assertRaises(TypeError):
            foo({23: 12.34, 123.456: 56.78})

        @arguments.accept({str: int})
        def bar(a):
            for k, v in a.items():
                self.assertIsInstance(k, str)
                self.assertIsInstance(v, int)

        bar({})
        bar({'foo': 23, 'bar': 42})

        with self.assertRaises(TypeError):
            bar()
        with self.assertRaises(TypeError):
            bar({'foo': 23, 'bar': 123.456})
        with self.assertRaises(TypeError):
            bar({'foo': 23, 666: 42})

    def test_add_varargs(self):
        @arguments.accept(arguments.sequenceof(int), add_varargs=True)
        def foo(a):
            self.assertTrue(misc.issequenceof(a, int))

        foo(123)
        foo(123, 456, 789)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123, 456, 'blah')
        with self.assertRaises(TypeError):
            foo(123, [456, 789])

        @arguments.accept(int, [int], add_varargs=True)
        def bar(a, b):
            self.assertEqual(a, 123)
            self.assertTupleEqual(b, (456, 789))

        bar(123, 456, 789)

    def test_flatten(self):
        @arguments.accept(arguments.flatten(int))
        def foo(a):
            self.assertListEqual(a, [123, 456, 789])

        foo([123, 456, 789])
        foo([123, [456, 789]])
        foo([123, [[456], 789]])
        foo([[[123, [[[[456]]], 789]]]])

        with self.assertRaises(TypeError):
            foo([123, ['blah', 789]])

    def test_each(self):
        @arguments.accept(arguments.each(
            int, arguments.condition(lambda x: x > 0)))
        def foo(a): pass

        foo(1)
        foo(123)

        with self.assertRaises(ValueError):
            foo(0)
        with self.assertRaises(ValueError):
            foo(-456)

    def test_either(self):
        @arguments.accept(arguments.either(int, str))
        def foo(a): pass

        foo(123)
        foo('blah')

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(123.456)

        @arguments.accept(arguments.either(int, arguments.sequenceof(str)))
        def bar(a): pass

        bar(123)
        bar(['doo', 'bee', 'doo'])

        with self.assertRaises(TypeError):
            bar('blah')
        with self.assertRaises(TypeError):
            bar([123, 456, 789])

    def test_condition(self):
        @arguments.accept(arguments.condition(lambda x: x%2 == 0))
        def foo(a): pass

        foo(2)
        foo(42)

        with self.assertRaises(ValueError):
            foo(1)
        with self.assertRaises(ValueError):
            foo(23)

        @arguments.accept(arguments.condition(lambda x: x > 0))
        def bar(a): pass

        bar(1)
        bar(42)

        with self.assertRaises(ValueError):
            bar(0)
        with self.assertRaises(ValueError):
            bar(-23)

    def test_repr(self):
        self.assertEqual(
            repr(arguments._make_constraint(int)),
            'int')
        self.assertEqual(
            repr(arguments._make_constraint(arguments.nullable(int))),
            'nullable(int)')
        self.assertEqual(
            repr(arguments._make_constraint([int])),
            '[int]')
        self.assertEqual(
            repr(arguments._make_constraint((23, 42))),
            '(23, 42)')
        self.assertEqual(
            repr(arguments._make_constraint((int, float, str))),
            '(int, float, str)')
        self.assertEqual(
            repr(arguments._make_constraint([int, float, str])),
            '[int, float, str]')
        self.assertEqual(
            repr(arguments._make_constraint({int: str})),
            '{int: str}')
        self.assertEqual(
            repr(arguments._make_constraint(arguments.flatten(int))),
            'flatten(int)')
        self.assertEqual(
            repr(arguments._make_constraint(arguments.each(int, float))),
            'each(int, float)')
        self.assertEqual(
            repr(arguments._make_constraint(arguments.either(int, str))),
            'either(int, str)')
        self.assertEqual(
            repr(arguments._make_constraint(lambda x: x / 2)),
            'lambda x: x / 2')
        self.assertEqual(
            repr(arguments._make_constraint(
                arguments.condition(lambda x: x < 3))),
            'condition(lambda x: x < 3)')

        def foo(x): x

        constraint = arguments.either(
            arguments.each(int, arguments.condition(lambda x: x%2 == 0)),
            arguments.sequenceof(foo),
            str,
        )
        reprs = 'either(each(int, condition(lambda x: x%2 == 0)), [foo], str)'

        self.assertEqual(repr(arguments._make_constraint(constraint)), reprs)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class BaseTestCase(MididingsTestCase):

    def test_Pass(self):
        self.check_patch(Pass(), {
            self.make_event(): True,
        })
        self.check_patch(Discard(), {
            self.make_event(): False,
        })

    def test_Fork(self):
        ev = self.make_event(port=2, channel=7)

        p = Fork([Pass(), Pass()])
        self.check_patch(p, {ev: True})

        p = Pass() // Pass()
        self.check_patch(p, {ev: True})

        p = +Port(3)
        self.check_patch(p, {ev: [ev, self.modify_event(ev, port=3)]})

        p = Channel(12).add()
        self.check_patch(p, {ev: [ev, self.modify_event(ev, channel=12)]})

        p = Fork([Pass(), Discard(), Pass()])
        self.check_patch(p, {ev: [ev]})

        p = Fork([Pass(), Pass()], remove_duplicates=False)
        self.check_patch(p, {ev: [ev, ev]})

        p = Fork([Pass(), Discard(), Pass()], remove_duplicates=False)
        self.check_patch(p, {ev: [ev, ev]})

    def test_Filter(self):
        self.check_filter(Filter([PROGRAM]), {
            self.make_event(NOTEON): (False, True),
            self.make_event(NOTEOFF): (False, True),
            self.make_event(CTRL): (False, True),
            self.make_event(PROGRAM): (True, False),
        })

        self.check_filter(Filter(types=NOTE), {
            self.make_event(NOTEON): (True, False),
            self.make_event(NOTEOFF): (True, False),
            self.make_event(CTRL): (False, True),
            self.make_event(PROGRAM): (False, True),
        })

        self.check_filter(Filter(NOTE, CTRL, AFTERTOUCH), {
            self.make_event(NOTEON): (True, False),
            self.make_event(NOTEOFF): (True, False),
            self.make_event(CTRL): (True, False),
            self.make_event(PROGRAM): (False, True),
        })

        self.check_filter(
                Filter(types=[CTRL, [PROGRAM], AFTERTOUCH|PITCHBEND]), {
            self.make_event(NOTEON): (False, True),
            self.make_event(CTRL): (True, False),
            self.make_event(PROGRAM): (True, False),
            self.make_event(AFTERTOUCH): (True, False),
            self.make_event(PITCHBEND): (True, False),
        })

    def test_Type(self):
        self.check_patch(NOTE, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL): False,
        })

        self.check_patch(~CTRL, {
            self.make_event(CTRL): False,
            self.make_event(NOTEON): True,
            self.make_event(PROGRAM): True,
        })

        self.check_patch(~(NOTE|AFTERTOUCH), {
            self.make_event(CTRL): True,
            self.make_event(NOTEON): False,
            self.make_event(AFTERTOUCH): False,
            self.make_event(PROGRAM): True,
        })

    def test_Split(self):
        ev1 = self.make_event(NOTEON, channel=1)
        ev2 = self.make_event(PROGRAM, channel=2)
        ev3 = self.make_event(CTRL)

        p = Split({ NOTE: Channel(4), PROGRAM: Channel(8) })
        self.check_patch(p, {
            ev1: [self.modify_event(ev1, channel=4)],
            ev2: [self.modify_event(ev2, channel=8)],
            ev3: [],
        })

    def test_Selector(self):
        p = CtrlFilter(23) % CtrlValueFilter(123)
        self.check_patch(p, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL, ctrl=23, value=42): False,
            self.make_event(CTRL, ctrl=42, value=123): True,
        })

        p = CtrlFilter(42) % CtrlValueFilter(123)
        self.check_patch(p, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL, ctrl=23, value=42): True,
            self.make_event(CTRL, ctrl=42, value=123): True,
        })

        p = (Filter(CTRL) & CtrlFilter(42) & CtrlValueFilter(123)) % Discard()
        self.check_patch(p, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL, ctrl=23, value=42): True,
            self.make_event(CTRL, ctrl=42, value=123): False,
        })

        p = (CtrlFilter(42) | CtrlValueFilter(123)) % Discard()
        self.check_patch(p, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL, ctrl=23, value=42): True,
            self.make_event(CTRL, ctrl=42, value=123): False,
        })

        p = (Filter(NOTE) |
             (CtrlFilter(42) & CtrlValueFilter(123))) % Discard()
        self.check_patch(p, {
            self.make_event(NOTEON): False,
            self.make_event(CTRL, ctrl=23, value=42): True,
            self.make_event(CTRL, ctrl=42, value=123): False,
        })

        p = CtrlFilter(42) % (CtrlValueFilter(123) % Discard())
        self.check_patch(p, {
            self.make_event(NOTEON): True,
            self.make_event(CTRL, ctrl=23, value=42): True,
            self.make_event(CTRL, ctrl=42, value=123): False,
        })

    def test_SelectorElse(self):
        p = PortFilter(2) % (Port(3), Port(4))
        ev1 = self.make_event(port=1)
        ev2 = self.make_event(port=2)
        self.check_patch(p, {
            ev1: [self.modify_event(ev1, port=4)],
            ev2: [self.modify_event(ev2, port=3)],
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class SetupTestCase(MididingsTestCase):

    def test_config_backend(self):
        config(backend='alsa')
        config(backend='jack')
        config(backend='jack-rt')
        with self.assertRaises(ValueError):
            config(backend='unknown')
        with self.assertRaises(ValueError):
            config(backend=666)

    def test_config_client_name(self):
        config(client_name='foo')
        with self.assertRaises(TypeError):
            config(client_name=666)

    def test_config_in_ports(self):
        config(in_ports=23)
        config(in_ports=['foo', 'bar'])
        config(in_ports=['foo', ['bar', 'blah']])
        config(in_ports=['foo', ('bar', 'blah', 'blubb')])
        with self.assertRaises(TypeError):
            config(in_ports='foo')
        with self.assertRaises(TypeError):
            config(in_ports=[23, 42])
#        with self.assertRaises(ValueError):
        with self.assertRaises(TypeError):
            config(in_ports=-1)

    def test_config_out_ports(self):
        config(out_ports=23)
        config(out_ports=['foo', 'bar'])
        config(out_ports=['foo', ['bar', 'blah']])
        config(out_ports=['foo', ('bar', 'blah', 'blubb')])
        with self.assertRaises(TypeError):
            config(out_ports='foo')
        with self.assertRaises(TypeError):
            config(out_ports=[23, 42])
#        with self.assertRaises(ValueError):
        with self.assertRaises(TypeError):
            config(out_ports=-1)

    def test_config_data_offset(self):
        config(data_offset=0)
        config(data_offset=1)
        with self.assertRaises(ValueError):
            config(data_offset=-1)
        with self.assertRaises(ValueError):
            config(data_offset=2)

    def test_config_octave_offset(self):
        for x in range(-3, 5):
            config(octave_offset=x)
        with self.assertRaises(TypeError):
            config(octave_offset=1.5)

    def test_config_initial_scene(self):
        config(initial_scene=4)
        config(initial_scene=(8,))
        config(initial_scene=(15,16))

    def test_config_start_delay(self):
        config(start_delay=None)
        config(start_delay=0)
        config(start_delay=1.23)

    @data_offsets
    def test_named_ports(self, off):
        config(out_ports = ['foo', 'bar', 'baz'])
        ev = self.make_event(port=off(0))

        self.check_patch(Port('bar'), {
            ev : [self.modify_event(ev, port=off(1))],
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class GeneratorsTestCase(MididingsTestCase):

    @data_offsets
    def test_NoteOn(self, off):
        ev = self.make_event(CTRL)

        p = NoteOn(60, 127)
        self.check_patch(p, {
            ev: [self.make_event(NOTEON, ev.port, ev.channel, 60, 127)],
        })
        p = NoteOn(port=off(2), channel=off(3), note=EVENT_CTRL, velocity=42)
        self.check_patch(p, {
            ev: [self.make_event(NOTEON, off(2), off(3), ev.ctrl, 42)],
        })

    @data_offsets
    def test_NoteOff(self, off):
        ev = self.make_event(CTRL)

        p = NoteOff(60, 127)
        self.check_patch(p, {
            ev: [self.make_event(NOTEOFF, ev.port, ev.channel, 60, 127)],
        })
        p = NoteOff(port=off(2), channel=off(3), note=EVENT_CTRL, velocity=42)
        self.check_patch(p, {
            ev: [self.make_event(NOTEOFF, off(2), off(3), ev.ctrl, 42)],
        })

    @data_offsets
    def test_Ctrl(self, off):
        ev = self.make_event(NOTEON)

        p = Ctrl(23, 42)
        self.check_patch(p, {
            ev: [self.make_event(CTRL, ev.port, ev.channel, 23, 42)],
        })
        p = Ctrl(port=off(2), channel=off(3), ctrl=23, value=EVENT_NOTE)
        self.check_patch(p, {
            ev: [self.make_event(CTRL, off(2), off(3), 23, ev.note)],
        })

    @data_offsets
    def test_Pitchbend(self, off):
        ev = self.make_event(CTRL)

        p = Pitchbend(8191)
        self.check_patch(p, {
            ev: [self.make_event(PITCHBEND, ev.port, ev.channel, 0, 8191)],
        })
        p = Pitchbend(port=off(2), channel=off(3), value=EVENT_VALUE)
        self.check_patch(p, {
            ev: [self.make_event(PITCHBEND, off(2), off(3), 0, ev.value)],
        })

    @data_offsets
    def test_Aftertouch(self, off):
        ev = self.make_event(CTRL)

        p = Aftertouch(42)
        self.check_patch(p, {
            ev: [self.make_event(AFTERTOUCH, ev.port, ev.channel, 0, 42)],
        })
        p = Aftertouch(port=off(2), channel=off(3), value=EVENT_VALUE)
        self.check_patch(p, {
            ev: [self.make_event(AFTERTOUCH, off(2), off(3), 0, ev.value)],
        })

    @data_offsets
    def test_Program(self, off):
        ev = self.make_event(NOTEON)

        p = Program(off(42))
        self.check_patch(p, {
            ev: [self.make_event(PROGRAM, ev.port, ev.channel, 0, 42)],
        })
        p = Program(port=off(2), channel=off(3), program=EVENT_NOTE)
        self.check_patch(p, {
            ev: [self.make_event(PROGRAM, off(2), off(3), 0, ev.note)],
        })

    @data_offsets
    def test_SysEx(self, off):
        ev = self.make_event(NOTEON)

        p = SysEx([0xf0, 4, 8, 15, 16, 23, 42, 0xf7])
        self.check_patch(p, {
            ev: [SysExEvent(ev.port, [0xf0, 4, 8, 15, 16, 23, 42, 0xf7])],
        })
        p = SysEx(port=off(2), sysex=[0xf0, 4, 8, 15, 16, 23, 42, 0xf7])
        self.check_patch(p, {
            ev: [SysExEvent(off(2), [0xf0, 4, 8, 15, 16, 23, 42, 0xf7])],
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class FiltersTestCase(MididingsTestCase):

    @data_offsets
    def test_PortFilter(self, off):
        config(in_ports = ['foo', 'bar', 'baz'])

        self.check_filter(PortFilter(off(0)), {
            self.make_event(port=off(0)): (True, False),
            self.make_event(port=off(1)): (False, True),
        })

        self.check_filter(PortFilter(ports=off(1)), {
            self.make_event(port=off(0)): (False, True),
            self.make_event(port=off(1)): (True, False),
        })

        self.check_filter(PortFilter(off(0), off(1), off(2)), {
            self.make_event(port=off(0)): (True, False),
            self.make_event(port=off(1)): (True, False),
            self.make_event(port=off(2)): (True, False),
            self.make_event(port=off(3)): (False, True),
        })

        self.check_filter(PortFilter(ports=['foo', 'baz']), {
            self.make_event(port=off(0)): (True, False),
            self.make_event(port=off(1)): (False, True),
            self.make_event(port=off(2)): (True, False),
            self.make_event(port=off(3)): (False, True),
        })

        with self.assertRaises(TypeError):
            PortFilter()
        with self.assertRaises(TypeError):
            PortFilter(123.456)
        with self.assertRaises(ValueError):
            PortFilter('nonexist')
        with self.assertRaises(ValueError):
            PortFilter(off(-1))

    @data_offsets
    def test_ChannelFilter(self, off):
        self.check_filter(ChannelFilter(off(2)), {
            self.make_event(channel=off(2)): (True, False),
            self.make_event(channel=off(3)): (False, True),
        })

        self.check_filter(ChannelFilter(channels=[off(3)]), {
            self.make_event(channel=off(2)): (False, True),
            self.make_event(channel=off(3)): (True, False),
        })

        self.check_filter(ChannelFilter(off(3), off(4), off(5)), {
            self.make_event(channel=off(2)): (False, True),
            self.make_event(channel=off(3)): (True, False),
            self.make_event(channel=off(4)): (True, False),
            self.make_event(channel=off(5)): (True, False),
        })

        self.check_filter(ChannelFilter([off(3), off(4), off(5)]), {
            self.make_event(channel=off(2)): (False, True),
            self.make_event(channel=off(3)): (True, False),
            self.make_event(channel=off(4)): (True, False),
            self.make_event(channel=off(5)): (True, False),
        })

        with self.assertRaises(TypeError):
            ChannelFilter()
        with self.assertRaises(TypeError):
            ChannelFilter('blah')
        with self.assertRaises(ValueError):
            ChannelFilter(off(-1))
        with self.assertRaises(ValueError):
            ChannelFilter(off(16))

    @data_offsets
    def test_KeyFilter(self, off):
        self.check_filter(KeyFilter('e2:a3'), {
            self.make_event(NOTE, note=51): (False, True),
            self.make_event(NOTE, note=52): (True, False),
            self.make_event(NOTE, note=68): (True, False),
            self.make_event(NOTE, note=69): (False, True),
            self.make_event(PROGRAM): (True, True),
            self.make_event(CTRL): (True, True),
        })

        self.check_filter(KeyFilter('c3'), {
            self.make_event(NOTE, note=59): (False, True),
            self.make_event(NOTE, note=60): (True, False),
            self.make_event(NOTE, note=61): (False, True),
        })

        self.check_filter(KeyFilter('e2', 'a3'), {
            self.make_event(NOTE, note=23): (False, True),
            self.make_event(NOTE, note=60): (True, False),
        })

        self.check_filter(KeyFilter(lower='c3'), {
            self.make_event(NOTE, note=23): (False, True),
            self.make_event(NOTE, note=59): (False, True),
            self.make_event(NOTE, note=60): (True, False),
            self.make_event(NOTE, note=108): (True, False),
        })

        self.check_filter(KeyFilter(upper=60), {
            self.make_event(NOTE, note=23): (True, False),
            self.make_event(NOTE, note=59): (True, False),
            self.make_event(NOTE, note=60): (False, True),
            self.make_event(NOTE, note=108): (False, True),
        })

        self.check_filter(KeyFilter(notes=[23, 42]), {
            self.make_event(NOTE, note=16): (False, True),
            self.make_event(NOTE, note=23): (True, False),
            self.make_event(NOTE, note=32): (False, True),
            self.make_event(NOTE, note=42): (True, False),
            self.make_event(NOTE, note=69): (False, True),
        })

        with self.assertRaises(TypeError):
            KeyFilter()
        with self.assertRaises(TypeError):
            KeyFilter(123.456)

    @data_offsets
    def test_VelocityFilter(self, off):
        self.check_filter(VelocityFilter(42), {
            self.make_event(NOTEON, velocity=23): (False, True),
            self.make_event(NOTEON, velocity=42): (True, False),
            self.make_event(NOTEOFF, velocity=23): (True, True),
            self.make_event(NOTEOFF, velocity=42): (True, True),
            self.make_event(PROGRAM): (True, True),
            self.make_event(CTRL): (True, True),
        })

        self.check_filter(VelocityFilter(lower=42), {
            self.make_event(NOTEON, velocity=41): (False, True),
            self.make_event(NOTEON, velocity=42): (True, False),
            self.make_event(NOTEON, velocity=69): (True, False),
        })

        self.check_filter(VelocityFilter(upper=108), {
            self.make_event(NOTEON, velocity=23): (True, False),
            self.make_event(NOTEON, velocity=107): (True, False),
            self.make_event(NOTEON, velocity=108): (False, True),
        })

        self.check_filter(VelocityFilter(lower=64, upper=128), {
            self.make_event(NOTEON, velocity=23): (False, True),
            self.make_event(NOTEON, velocity=127): (True, False),
            self.make_event(NOTEOFF, velocity=23): (True, True),
            self.make_event(NOTEOFF, velocity=127): (True, True),
            self.make_event(PROGRAM): (True, True),
            self.make_event(CTRL): (True, True),
        })

        with self.assertRaises(TypeError):
            VelocityFilter()
        with self.assertRaises(TypeError):
            VelocityFilter(123.456)
        with self.assertRaises(TypeError):
            VelocityFilter('blah')
        with self.assertRaises(ValueError):
            VelocityFilter(-1)
        with self.assertRaises(ValueError):
            VelocityFilter(128)

    @data_offsets
    def test_CtrlFilter(self, off):
        self.check_filter(CtrlFilter(23), {
            self.make_event(CTRL, ctrl=23): (True, False),
            self.make_event(CTRL, ctrl=42): (False, True),
            self.make_event(NOTEON): (False, False),
            self.make_event(PROGRAM): (False, False),
        })

        self.check_filter(CtrlFilter(23, 42), {
            self.make_event(CTRL, ctrl=23): (True, False),
            self.make_event(CTRL, ctrl=42): (True, False),
            self.make_event(CTRL, ctrl=69): (False, True),
        })

        self.check_filter(CtrlFilter(ctrls=[23, 42]), {
            self.make_event(CTRL, ctrl=23): (True, False),
            self.make_event(CTRL, ctrl=42): (True, False),
            self.make_event(CTRL, ctrl=69): (False, True),
        })

        with self.assertRaises(TypeError):
            CtrlFilter(123.456)
        with self.assertRaises(TypeError):
            CtrlFilter('blah')
        with self.assertRaises(ValueError):
            CtrlFilter(-1)
        with self.assertRaises(ValueError):
            CtrlFilter(128)

    @data_offsets
    def test_CtrlValueFilter(self, off):
        self.check_filter(CtrlValueFilter(value=23), {
            self.make_event(CTRL, value=16): (False, True),
            self.make_event(CTRL, value=23): (True, False),
            self.make_event(CTRL, value=66): (False, True),
            self.make_event(NOTE): (False, False),
            self.make_event(PROGRAM): (False, False),
        })

        self.check_filter(CtrlValueFilter(23, 42), {
            self.make_event(CTRL, value=32): (True, False),
            self.make_event(CTRL, value=66): (False, True),
        })

    @data_offsets
    def test_ProgramFilter(self, off):
        self.check_filter(ProgramFilter(off(1)), {
            self.make_event(PROGRAM, program=off(1)): (True, False),
            self.make_event(PROGRAM, program=off(2)): (False, True),
            self.make_event(NOTE): (False, False),
            self.make_event(CTRL): (False, False),
        })

        self.check_filter(ProgramFilter(off(4), off(8), off(15), off(16)), {
            self.make_event(PROGRAM, program=off(8)): (True, False),
            self.make_event(PROGRAM, program=off(13)): (False, True),
        })

        self.check_filter(ProgramFilter(
                programs=[off(4), off(8), off(15), off(16)]), {
            self.make_event(PROGRAM, program=off(8)): (True, False),
            self.make_event(PROGRAM, program=off(13)): (False, True),
        })

        with self.assertRaises(TypeError):
            ProgramFilter()
        with self.assertRaises(TypeError):
            ProgramFilter(123.456)
        with self.assertRaises(TypeError):
            ProgramFilter('blah')
        with self.assertRaises(ValueError):
            ProgramFilter(off(-1))
        with self.assertRaises(ValueError):
            ProgramFilter(off(128))

    @data_offsets
    def test_SysExFilter(self, off):
        self.check_filter(
                SysExFilter([0xf0, 4, 8, 15, 16, 23, 42, 0xf7]), {
            SysExEvent(off(0), [0xf0, 4, 8, 15, 16, 23, 42, 0xf7]):
                (True, False),
            SysExEvent(off(0), [0xf0, 4, 8, 15, 16, 23, 43, 0xf7]):
                (False, True),
            self.make_event(NOTE):
                (False, False),
            self.make_event(CTRL):
                (False, False),
        })

        self.check_filter(SysExFilter('\xf0\x04\x08'), {
            SysExEvent(off(0), [0xf0, 4, 8, 15, 16, 23, 42, 0xf7]):
                (True, False),
            SysExEvent(off(0), [0xf0, 4, 9, 15, 16, 23, 42, 0xf7]):
                (False, True),
        })

        self.check_filter(SysExFilter(manufacturer=0x42), {
            SysExEvent(off(0), [0xf0, 0x42, 8, 15, 16, 23, 42, 0xf7]):
                (True, False),
            SysExEvent(off(0), [0xf0, 0x43, 8, 15, 16, 23, 42, 0xf7]):
                (False, True),
        })

        self.check_filter(SysExFilter(manufacturer='\x00\x23\x42'), {
            SysExEvent(off(0), [0xf0, 0x00, 0x23, 0x42, 16, 23, 42, 0xf7]):
                (True, False),
            SysExEvent(off(0), [0xf0, 4, 8, 15, 16, 23, 42, 0xf7]):
                (False, True),
        })



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

try:
    import unittest2 as unittest
except:
    import unittest

from mididings import overload


class OverloadTestCase(unittest.TestCase):

    def setUp(self):
        # yuck!
        self.registry_bak = overload._registry.copy()

    def tearDown(self):
        # yuck!
        overload._registry = self.registry_bak

    def test_call(self):
        funcs = [
            lambda a: 1,
            lambda a, b: 2,
        ]

        self.assertEqual(overload.call([23], {}, funcs), 1)
        self.assertEqual(overload.call([], {'a': 23}, funcs), 1)
        self.assertEqual(overload.call([23, 42], {}, funcs), 2)
        self.assertEqual(overload.call([23], {'b': 42}, funcs), 2)
        self.assertEqual(overload.call([], {'a': 23, 'b': 42}, funcs), 2)

        with self.assertRaises(TypeError):
            overload.call([], {}, funcs)
        with self.assertRaises(TypeError):
            overload.call([23, 42, 69], {}, funcs)
        with self.assertRaises(TypeError):
            overload.call([23, 42], {'c': 69}, funcs)

    def test_mark(self):
        @overload.mark
        def foo(a):
            return 1

        @overload.mark
        def foo(a, b):
            return 2

        self.assertEqual(foo(23), 1)
        self.assertEqual(foo(a=23), 1)
        self.assertEqual(foo(23, 42), 2)
        self.assertEqual(foo(a=23, b=42), 2)

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(23, 42, 69)
        with self.assertRaises(TypeError):
            foo(23, 42, c=69)

    def test_argnames(self):
        @overload.mark
        def foo(a, b):
            return 1

        @overload.mark
        def foo(c, d):
            return 2

        self.assertEqual(foo(23, 42), 1)
        self.assertEqual(foo(23, b=42), 1)
        self.assertEqual(foo(a=23, b=42), 1)
        self.assertEqual(foo(23, d=42), 2)
        self.assertEqual(foo(c=23, d=42), 2)

        with self.assertRaises(TypeError):
            foo(23, e=42)

    def test_defargs(self):
        @overload.mark
        def foo(a=23):
            return 1

        @overload.mark
        def foo(a, b=42):
            return 2

        @overload.mark
        def foo(a, b, c=69):
            return 3

        @overload.mark
        def foo(a, b=42, d=123):
            return 4

        self.assertEqual(foo(), 1)
        self.assertEqual(foo(23), 1)
        self.assertEqual(foo(a=23), 1)
        self.assertEqual(foo(23, 42), 2)
        self.assertEqual(foo(a=23, b=42), 2)
        self.assertEqual(foo(23, 42, 69), 3)
        self.assertEqual(foo(23, 42, c=69), 3)
        self.assertEqual(foo(a=23, b=42, c=69), 3)
        self.assertEqual(foo(23, 42, d=123), 4)
        self.assertEqual(foo(a=23, b=42, d=123), 4)
        self.assertEqual(foo(a=23, d=123), 4)

        with self.assertRaises(TypeError):
            foo(23, c=69)

    def test_varargs(self):
        @overload.mark
        def foo(a):
            return 1

        @overload.mark
        def foo(b, c, d=69):
            return 2

        @overload.mark
        def foo(d, e=42, *args):
            return 3

        self.assertEqual(foo(23), 1)
        self.assertEqual(foo(23, 42), 2)
        self.assertEqual(foo(23, 42, 69), 2)
        self.assertEqual(foo(23, 42, d=69), 2)
        self.assertEqual(foo(23, e=42), 3)
        self.assertEqual(foo(d=23, e=42), 3)
        self.assertEqual(foo(23, 42, 69, 123), 3)

        with self.assertRaises(TypeError):
            foo(e=23)

    def test_partial(self):
        @overload.partial((23,))
        def foo(a, b):
            return a, b

        self.assertEqual(foo(42), (23, 42))
        self.assertEqual(foo(23, 42), (23, 42))
        self.assertEqual(foo(b=42), (23, 42))
        self.assertEqual(foo(a=23, b=42), (23, 42))

        with self.assertRaises(TypeError):
            foo()
        with self.assertRaises(TypeError):
            foo(23, 42, 69)

        @overload.partial((23,))
        def bar(a, b, c=123):
            return a, b, c

        self.assertEqual(bar(42), (23, 42, 123))
        self.assertEqual(bar(23, 42), (23, 42, 123))
        self.assertEqual(bar(42, c=123), (23, 42, 123))
        self.assertEqual(bar(23, 42, 123), (23, 42, 123))

        with self.assertRaises(TypeError):
            bar()
        with self.assertRaises(TypeError):
            bar(23, 42, 69, 123)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *
from mididings.util import *


class UtilTestCase(MididingsTestCase):

    def test_note_number(self):
        config(octave_offset=1)
        self.assertEqual(note_number(42), 42)
        self.assertEqual(note_number('c4'), 60)
        self.assertEqual(note_number('c#4'), 61)
        self.assertEqual(note_number('db4'), 61)
        self.assertEqual(note_number('a0'), 21)

        with self.assertRaises(ValueError):
            note_number(-23)
        with self.assertRaises(TypeError):
            note_number(123.456)
        with self.assertRaises(ValueError):
            note_number('a23')
        with self.assertRaises(ValueError):
            note_number('h3')

        config(octave_offset=2)
        self.assertEqual(note_number(42), 42)
        self.assertEqual(note_number('c3'), 60)
        self.assertEqual(note_number('c#3'), 61)
        self.assertEqual(note_number('db3'), 61)
        self.assertEqual(note_number('a-1'), 21)

    def test_note_range(self):
        self.assertEqual(note_range(23), (23, 24))
        self.assertEqual(note_range((23, 42)), (23, 42))
        self.assertEqual(note_range('c3:c5'), (60, 84))
        self.assertEqual(note_range(':c4'), (0, 72))
        self.assertEqual(note_range('c2:'), (48, 0))

        with self.assertRaises(ValueError):
            note_range('blah')
        with self.assertRaises(ValueError):
            note_range('x3:y5')
        with self.assertRaises(ValueError):
            note_range((23,))
        with self.assertRaises(ValueError):
            note_range(('c4',))

    def test_note_name(self):
        config(octave_offset=1)
        self.assertEqual(note_name(60), 'c4')
        self.assertEqual(note_name(61), 'c#4')
        self.assertEqual(note_name(0), 'c-1')
        self.assertEqual(note_name(-23), 'c#-3')
        self.assertEqual(note_name(127), 'g9')

        with self.assertRaises(TypeError):
            note_name(123.456)

    def test_event_type(self):
        self.assertEqual(event_type(NOTEON), NOTEON)
        self.assertEqual(event_type(PROGRAM), PROGRAM)

        with self.assertRaises(ValueError):
            event_type(NOTE)
        with self.assertRaises(ValueError):
            event_type(123)

    @data_offsets
    def test_port_number(self, off):
        config(
            in_ports=['foo', 'bar'],
            out_ports=['foo', 'blah', 'bar']
        )

        self.assertEqual(port_number(1), 1)
        self.assertEqual(port_number('foo'), off(0))
        self.assertEqual(port_number('blah'), off(1))

        with self.assertRaises(ValueError):
            port_number('bar')
        with self.assertRaises(TypeError):
            port_number(123.456)

    def test_sysex_data(self):
        self.assertEqual(
            sysex_data('\xf0\x04\x08\x15\x16\x23\x42\xf7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data([0xf0, 0x04, 0x08, 0x15, 0x16, 0x23, 0x42, 0xf7]),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('F0 04 08 15 16 23 42 F7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('F0,04,08,15,16,23,42,F7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('f0 04 08 15 16 23 42 f7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('f0,04,08,15,16,23,42,f7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('F0040815162342F7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('f0040815162342f7'),
            self.native_sysex('\xf0\x04\x08\x15\x16\x23\x42\xf7'))
        self.assertEqual(
            sysex_data('\xf0\x23\x42\x66', allow_partial=True),
            self.native_sysex('\xf0\x23\x42\x66'))

        with self.assertRaises(ValueError):
            sysex_data('F0')
        with self.assertRaises(ValueError):
            sysex_data('\xf0')
        with self.assertRaises(ValueError):
            sysex_data('\xfa\x23\x42\xf7')
        with self.assertRaises(ValueError):
            sysex_data('\xf0\x23\x42\xfa')
        with self.assertRaises(ValueError):
            sysex_data('\xf0\xff\xff\xfa')
        with self.assertRaises(ValueError):
            sysex_data('F0 FF FF FA')
        with self.assertRaises(ValueError):
            sysex_data('f0,23,42,fa')

    def test_sysex_manufacturer(self):
        self.assertEqual(sysex_manufacturer('\x42'),
                         self.native_sysex('\x42'))
        self.assertEqual(sysex_manufacturer(0x42),
                         self.native_sysex('\x42'))
        self.assertEqual(sysex_manufacturer('\x00\x42\x23'),
                         self.native_sysex('\x00\x42\x23'))
        self.assertEqual(sysex_manufacturer([0x00, 0x42, 0x23]),
                         self.native_sysex('\x00\x42\x23'))

        with self.assertRaises(ValueError):
            sysex_manufacturer('\x42\x23')
        with self.assertRaises(ValueError):
            sysex_manufacturer('\x42\x00\x23')
        with self.assertRaises(ValueError):
            sysex_manufacturer('\x00\x23\xff')



# -*- coding: utf-8 -*-
#
import sys, os

sys.path.insert(0, os.path.abspath('..'))

extensions = ['sphinx.ext.autodoc', 'sphinxcontrib.fulltoc']

templates_path = ['templates']
html_theme_path = ['theme']
exclude_patterns = ['build']

source_suffix = '.rst'
master_doc = 'index'

project = u'mididings'
copyright = u'2008-2014, Dominic Sacré'
version = '20230114'
release = ''

html_theme = 'nasophon'
html_copy_source = False
pygments_style = 'sphinx'

add_module_names = False
autodoc_member_order = 'bysource'
autodoc_default_flags = ['members', 'undoc-members']


from sphinx.ext.autodoc import py_ext_sig_re
from sphinx.util.docstrings import prepare_docstring
#from sphinx.domains.python import PyModulelevel
from sphinx.domains.python import PyFunction
from sphinx import addnodes
import re


def process_docstring(app, what, name, obj, options, lines):
    """
    Remove leading function signatures from docstring.
    """
    while len(lines) and py_ext_sig_re.match(lines[0]) is not None:
        del lines[0]

def process_signature(app, what, name, obj,
                      options, signature, return_annotation):
    """
    Replace function signature with those specified in the docstring.
    """
    if hasattr(obj, '__doc__') and obj.__doc__ is not None:
        lines = prepare_docstring(obj.__doc__)
        siglines = []

        for line in lines:
            if py_ext_sig_re.match(line) is not None:
                siglines.append(line)
            else:
                break

        if len(siglines):
            siglines[0] = siglines[0][siglines[0].index('('):]
            return ('\n'.join(siglines), None)

    return (signature, return_annotation)


class DingsFunction(PyFunction):
    """
    Stripped-down version of the 'function::' directive that accepts an
    additional argument in angle brackets, used to specify the node's
    'fullname' attribute. This allows proper cross-references to mididings
    operators.
    """
    def handle_signature(self, sig, signode):
        m = re.match('(.*) <([\w.]*)>', sig)
        if m:
            op = m.group(1)
            name = m.group(2)
            modname = self.options.get('module',
                                       self.env.temp_data.get('py:module'))

            signode['module'] = modname
            signode['class'] = ''
            signode['fullname'] = name
            signode += addnodes.desc_name(op, op)

            return name, None
        else:
            return super(PyFunction, self).handle_signature(sig, signode)


def setup(app):
    app.connect('autodoc-process-docstring', process_docstring)
    app.connect('autodoc-process-signature', process_signature)
    app.add_directive('dingsfun', DingsFunction)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *
from mididings.event import *
import mididings.constants

import copy
import pickle


class EventTestCase(MididingsTestCase):

    @data_offsets
    def test_NoteOnEvent(self, off):
        ev = NoteOnEvent(off(1), off(7), 60, 127)

        def foo(ev):
            self.assertEqual(ev.type, NOTEON)
            self.assertEqual(ev.port, off(1))
            self.assertEqual(ev.port_, 1)
            self.assertEqual(ev.channel, off(7))
            self.assertEqual(ev.channel_, 7)
            self.assertEqual(ev.note, 60)
            self.assertEqual(ev.data1, 60)
            self.assertEqual(ev.velocity, 127)
            self.assertEqual(ev.data2, 127)
            ev.port = off(3)
            ev.channel = off(5)
            ev.note = 69
            ev.velocity = 42
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.type, NOTEON)
        self.assertEqual(r.port, off(3))
        self.assertEqual(r.port_, 3)
        self.assertEqual(r.channel, off(5))
        self.assertEqual(r.channel_, 5)
        self.assertEqual(r.note, 69)
        self.assertEqual(r.data1, 69)
        self.assertEqual(r.velocity, 42)
        self.assertEqual(r.data2, 42)
        with self.assertRaises(AttributeError): r.ctrl
        with self.assertRaises(AttributeError): r.value
        with self.assertRaises(AttributeError): r.program
        with self.assertRaises(AttributeError): r.sysex

    @data_offsets
    def test_NoteOffEvent(self, off):
        ev = NoteOffEvent(off(1), off(7), 60, 127)

        def foo(ev):
            self.assertEqual(ev.type, NOTEOFF)
            self.assertEqual(ev.port, off(1))
            self.assertEqual(ev.port_, 1)
            self.assertEqual(ev.channel, off(7))
            self.assertEqual(ev.channel_, 7)
            self.assertEqual(ev.note, 60)
            self.assertEqual(ev.data1, 60)
            self.assertEqual(ev.velocity, 127)
            self.assertEqual(ev.data2, 127)
            ev.port = off(3)
            ev.channel = off(5)
            ev.note = 69
            ev.velocity = 42
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.type, NOTEOFF)
        self.assertEqual(r.port, off(3))
        self.assertEqual(r.port_, 3)
        self.assertEqual(r.channel, off(5))
        self.assertEqual(r.channel_, 5)
        self.assertEqual(r.note, 69)
        self.assertEqual(r.data1, 69)
        self.assertEqual(r.velocity, 42)
        self.assertEqual(r.data2, 42)
        with self.assertRaises(AttributeError): r.ctrl
        with self.assertRaises(AttributeError): r.value
        with self.assertRaises(AttributeError): r.program
        with self.assertRaises(AttributeError): r.sysex

    @data_offsets
    def test_CtrlEvent(self, off):
        ev = CtrlEvent(off(1), off(7), 60, 127)

        def foo(ev):
            self.assertEqual(ev.type, CTRL)
            self.assertEqual(ev.port, off(1))
            self.assertEqual(ev.port_, 1)
            self.assertEqual(ev.channel, off(7))
            self.assertEqual(ev.channel_, 7)
            self.assertEqual(ev.ctrl, 60)
            self.assertEqual(ev.data1, 60)
            self.assertEqual(ev.value, 127)
            self.assertEqual(ev.data2, 127)
            ev.port = off(3)
            ev.channel = off(5)
            ev.ctrl = 69
            ev.value = 42
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.type, CTRL)
        self.assertEqual(r.port, off(3))
        self.assertEqual(r.port_, 3)
        self.assertEqual(r.channel, off(5))
        self.assertEqual(r.channel_, 5)
        self.assertEqual(r.ctrl, 69)
        self.assertEqual(r.data1, 69)
        self.assertEqual(r.value, 42)
        self.assertEqual(r.data2, 42)
        with self.assertRaises(AttributeError): r.note
        with self.assertRaises(AttributeError): r.velocity
        with self.assertRaises(AttributeError): r.program
        with self.assertRaises(AttributeError): r.sysex

    @data_offsets
    def test_ProgramEvent(self, off):
        ev = ProgramEvent(off(1), off(7), off(42))

        def foo(ev):
            self.assertEqual(ev.type, PROGRAM)
            self.assertEqual(ev.port, off(1))
            self.assertEqual(ev.port_, 1)
            self.assertEqual(ev.channel, off(7))
            self.assertEqual(ev.channel_, 7)
            self.assertEqual(ev.program, off(42))
            self.assertEqual(ev.data2, 42)
            ev.port = off(3)
            ev.channel = off(5)
            ev.program = off(66)
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.type, PROGRAM)
        self.assertEqual(r.port, off(3))
        self.assertEqual(r.port_, 3)
        self.assertEqual(r.channel, off(5))
        self.assertEqual(r.channel_, 5)
        self.assertEqual(r.program, off(66))
        self.assertEqual(r.data2, 66)
        with self.assertRaises(AttributeError): r.note
        with self.assertRaises(AttributeError): r.velocity
        with self.assertRaises(AttributeError): r.ctrl
        with self.assertRaises(AttributeError): r.value
        with self.assertRaises(AttributeError): r.sysex

    @data_offsets
    def test_SysExEvent(self, off):
        sysex1 = '\xf0\x04\x08\x15\x16\x23\x42\xf7'
        sysex2 = '\xf0\x09\x11\x02\x74\x5b\x41\x56\x63\x56\xf7'
        ev = SysExEvent(off(1), sysex1)

        def foo(ev):
            self.assertEqual(ev.type, SYSEX)
            self.assertEqual(ev.port, off(1))
            self.assertEqual(ev.sysex, self.native_sysex(sysex1))
            ev.port = off(3)
            ev.sysex = sysex2
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.type, SYSEX)
        self.assertEqual(r.port, off(3))
        self.assertEqual(r.port_, 3)
        self.assertEqual(r.sysex, self.native_sysex(sysex2))
        with self.assertRaises(AttributeError): ev.note
        with self.assertRaises(AttributeError): ev.velocity
        with self.assertRaises(AttributeError): ev.ctrl
        with self.assertRaises(AttributeError): ev.value
        with self.assertRaises(AttributeError): ev.program

    @data_offsets
    def test_SysExEvent_modify(self, off):
        sysex1 = '\xf0\x04\x08\x15\x16\x23\x42\xf7'
        sysex2 = '\xf0\x05\x08\x15\x16\x23\x42\xf7'
        ev = SysExEvent(off(1), sysex1)

        def foo(ev):
            ev.sysex[1] = 0x05
            return ev

        (r,) = self.run_patch(Process(foo), ev)

        self.assertEqual(r.sysex, self.native_sysex(sysex2))

    @data_offsets
    def test_operator_equals(self, off):
        for t in constants._EVENT_TYPES.values():
            a = self.make_event(type=t, port=off(0))
            b = self.make_event(type=t, port=off(1))
            c = self.make_event(type=a.type, port=a.port, channel=a.channel,
                                data1=a.data1, data2=a.data2, sysex=a.sysex_)

            self.assertNotEqual(a, b)
            self.assertFalse(a == b)
            self.assertTrue(a != b)

            self.assertEqual(a, c)
            self.assertTrue(a == c)
            self.assertFalse(a != c)

    @data_offsets
    def test_to_string(self, off):
        for t in constants._EVENT_TYPES.values():
            ev = self.make_event(t)
            self.assertIsInstance(ev.to_string(), str)
            self.assertIsInstance(ev.to_string(['foo', 'bar'], 3, 80), str)

    @data_offsets
    def test_rebuild_repr(self, off):
        for t in constants._EVENT_TYPES.values():
            ev = self.make_event(t)
            rebuilt = eval(repr(ev), self.mididings_dict)
            self.assertEqual(rebuilt, ev)
            self.assertEqual(repr(ev), repr(rebuilt))

    @data_offsets
    def test_copy(self, off):
        a = self.make_event()
        b = copy.copy(a)

        self.assertEqual(b, a)

        c = SysExEvent(off(23), '\xf0\x04\x08\x15\x16\x23\x42\xf7')
        d = copy.copy(c)

        self.assertEqual(d, c)

    @data_offsets
    def test_pickle(self, off):
        a = self.make_event()
        b = pickle.loads(pickle.dumps(a))

        self.assertEqual(b, a)

        c = SysExEvent(off(23), '\xf0\x04\x08\x15\x16\x23\x42\xf7')
        d = pickle.loads(pickle.dumps(c))

        self.assertEqual(d, c)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *
from mididings import engine


class EngineTestCase(MididingsTestCase):

    def test_in_ports(self):
        ports = ['foo', 'bar', 'baz']
        config(in_ports = ports)
        self.assertEqual(engine.in_ports(), ports)

        config(in_ports = 3)
        self.assertEqual(engine.in_ports(), ['in_0', 'in_1', 'in_2'])

    def test_out_ports(self):
        ports = ['foo', 'bar', 'baz']
        config(out_ports = ports)
        self.assertEqual(engine.out_ports(), ports)

        config(out_ports = 3)
        self.assertEqual(engine.out_ports(), ['out_0', 'out_1', 'out_2'])

    def test_active(self):
        self.assertFalse(engine.active())

        def foo(ev):
            self.assertTrue(engine.active())

        self.run_patch(Process(foo), self.make_event())



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

from tests.helpers import *

from mididings import *


class EngineTestCase(MididingsTestCase):

    @data_offsets
    def testSanitize(self, off):
        def foo(ev):
            ev.port = off(42)
        def bar(ev):
            self.fail()
        p = Process(foo) >> Sanitize() >> Process(bar)
        self.run_patch(p, self.make_event(port=off(42)))

        p = Velocity(+666) >> Sanitize()
        r = self.run_patch(p, self.make_event(NOTEON, velocity=42))
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0].data2, 127)

    @data_offsets
    def testSceneSwitch(self, off):
        config(silent=True)
        p = {
            off(0): Split({
                        PROGRAM:  SceneSwitch(),
                        ~PROGRAM: Channel(off(7)),
                    }),
            off(1): Channel(off(13)),
        }
        events = (
            self.make_event(NOTEON, off(0), off(0), 69, 123),
            self.make_event(PROGRAM, off(0), off(0), 0, 1),   # no data offset!
            self.make_event(NOTEON, off(0), off(0), 23, 42),
            self.make_event(NOTEOFF, off(0), off(0), 69, 0),
        )
        results = [
            self.make_event(NOTEON, off(0), off(7), 69, 123),
            self.make_event(NOTEON, off(0), off(13), 23, 42),
            self.make_event(NOTEOFF, off(0), off(7), 69, 0),
        ]
        self.check_scenes(p, {
            events: results,
        })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import platform
import sys
import sysconfig

from subprocess import getstatusoutput
from setuptools import setup, Extension


def extract_version():
    """
    Extract version from pyproject.toml.

    This exists so pyproject.toml is the single source of truth for the
    package version.
    """
    # initialise with placeholder
    version = "dev"

    # compile regexp
    regexp = re.compile(r"version = \"(?P<version>\d*)\"")

    # open pyproject.toml
    with open("pyproject.toml") as input_file:
        lines = input_file.readlines()

    # extract version
    for line in lines:
        if line.startswith("version"):
            match = regexp.match(line)

            if match:
                version = match.group("version")
                break

    return version


version = extract_version()

config = {
    'alsa-seq':     (platform.system() == 'Linux'),
    'jack-midi':    True,
    'debug':        False,
}

# information for compiled extension
sources = [
    "src/engine.cc",
    "src/patch.cc",
    "src/python_caller.cc",
    "src/send_midi.cc",
    "src/python_module.cc",
    "src/backend/base.cc",
]
library_dirs = []
include_dirs = ["src"]
libraries = []
define_macros = [("VERSION", f"{version}")]


def pkgconfig(name):
    """
    Run pkg-config for the given package, and add the required flags to our
    list of build arguments.
    """
    status, output = getstatusoutput(f"pkg-config --libs --cflags {name}")
    if status:
        sys.exit(f"couldn't find package '{name}'")
    for token in output.split():
        opt, val = token[:2], token[2:]
        if opt == "-I":
            include_dirs.append(val)
        elif opt == "-l":
            libraries.append(val)
        elif opt == "-L":
            library_dirs.append(val)


def library_path_dirs():
    """
    Extract library paths from LIBRARY_PATH environment variable.
    """
    library_path = os.environ.get("LIBRARY_PATH")
    return library_path.split(":") if library_path else []


def lib_dirs():
    """
    Attempt to return the compiler's library search paths.

    Also include paths from LIBRARY_PATH environment variable.
    """
    env_libs = library_path_dirs()
    try:
        status, output = getstatusoutput(
            "LC_ALL=C " + sysconfig.get_config_var("CC") + " -print-search-dirs"
        )
        for line in output.splitlines():
            if "libraries: =" in line:
                libdirs = [
                    os.path.normpath(p) for p in line.split("=", 1)[1].split(":")
                ]
                return env_libs + libdirs
        return env_libs
    except Exception:
        return env_libs


def boost_lib_name(name, add_suffixes=[]):
    """
    Try to figure out the correct boost library name (with or without "-mt"
    suffix, or with any of the given additional suffixes).
    """
    libdirs = lib_dirs()
    for suffix in add_suffixes + ["", "-mt"]:
        for libdir in libdirs:
            for ext in ["so"] + ["dylib"] * (sys.platform == "darwin"):
                libname = f"lib{name}{suffix}.{ext}"
                if os.path.isfile(os.path.join(libdir, libname)):
                    return name + suffix
    return name


libraries.append(
    boost_lib_name("boost_python", [f"{sys.version_info[0]}{sys.version_info[1]}"])
)
libraries.append(boost_lib_name("boost_thread"))

library_dirs.extend(library_path_dirs())

if config["alsa-seq"]:
    define_macros.append(("ENABLE_ALSA_SEQ", 1))
    sources.append("src/backend/alsa.cc")
    pkgconfig("alsa")

if config["jack-midi"]:
    define_macros.append(("ENABLE_JACK_MIDI", 1))
    sources.extend(
        [
            "src/backend/jack.cc",
            "src/backend/jack_buffered.cc",
            "src/backend/jack_realtime.cc",
        ]
    )
    pkgconfig("jack")

setup(
    ext_modules=[
        Extension(
            name="_mididings",
            sources=sources,
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=libraries,
            define_macros=define_macros,
        ),
    ],
    packages = [
        'mididings',
        'mididings.units',
        'mididings.extra',
        'mididings.live',
    ],
    scripts = [
        'scripts/mididings',
        'scripts/livedings',
        'scripts/send_midi',
    ],
    data_files=[
        (
            "share/man/man1",
            [
                "doc/man/mididings.1",
                "doc/man/livedings.1",
                "doc/man/send_midi.1",
            ],
        ),
    ],
)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

try:
    import unittest2 as unittest
except:
    import unittest

import random
import itertools
import sys
import copy

from mididings import *
from mididings import setup, engine, misc, constants
from mididings.event import *

import mididings
import mididings.event


def data_offsets(f):
    """
    Run the test f twice, with data offsets 0 and 1
    """
    def data_offset_wrapper(self):
        for offset in (0, 1):
            def off(n):
                return n + offset
            config(data_offset = offset)
            f(self, off)
    data_offset_wrapper.__name__ = f.__name__
    return data_offset_wrapper


class MididingsTestCase(unittest.TestCase):
    def setUp(self):
        setup.reset()
        setup.config(data_offset = 0)

        self.mididings_dict = mididings.__dict__.copy()
        self.mididings_dict.update(mididings.event.__dict__)

    def check_patch(self, patch, d):
        """
        Test the given patch. d must be a mapping from events to the expected
        list of resulting events.
        """
        self.check_scenes({ setup.get_config('data_offset'): patch }, d)

    def check_scenes(self, scenes, d):
        """
        Test the given scenes. d must be a mapping from events to the expected
        list of resulting events.
        """
        for ev, expected in d.items():
            r = list(itertools.chain(*self.run_scenes(scenes, ev)))
            if isinstance(expected, bool):
                # boolean value: ensure that at most one event was returned
                self.assertLessEqual(len(r), 1)
                # ensure that event is unchanged
                if len(r):
                    self.assertEqual(r[0], ev)
                # check if the result is as expected
                self.assertEqual(bool(len(r)), expected,
                        "\nscenes=%s\nev=%s\nexpected=%s" %
                            (repr(scenes), repr(ev), repr(expected)))
            else:
                # list: check if the result is exactly as expected
                self.assertEqual(r, expected,
                        "\nscenes=%s\nev=%s\nr=%s\nexpected=%s" %
                            (repr(scenes), repr(ev), repr(r), repr(expected)))

    def check_filter(self, filt, d):
        """
        Test if the filter filt works as expected.
        d must be a mapping from events to a tuple of two booleans, where the
        first value specifies if the event should match the filter as is, and
        the second value specifies if the event should match the inverted
        filter.
        """
        for ev, expected in d.items():
            for f, p in zip((filt, ~filt, -filt),
                            (expected[0], expected[1], not expected[0])):
                self.check_patch(f, {ev: p})

    def check_sequence(self, scenes, events):
        # TODO
        pass

    def run_patch(self, patch, events):
        """
        Run the given events through the given patch, return the list of
        resulting events.
        """
        r = self.run_scenes({ setup.get_config('data_offset'): patch }, events)
        return list(itertools.chain(*r))

    def run_scenes(self, scenes, events):
        """
        Run the given events through the given scenes, return the list of
        resulting events.
        """
        # run scenes
        r1 = self._run_scenes_impl(scenes, events)

        # check if events can be rebuilt from their repr()
        for r in r1:
            for ev in r:
                rebuilt = eval(repr(ev), self.mididings_dict)
                self.assertEqual(rebuilt, ev)

        rebuilt = self._rebuild_repr(scenes)
        if rebuilt is not None:
            # run scenes rebuilt from their repr(), result should be identical
            r2 = self._run_scenes_impl(rebuilt, events)
            self.assertEqual(r2, r1)

        return r1

    def _run_scenes_impl(self, scenes, events):
        # set up a dummy engine
        setup._config_impl(backend='dummy')
        e = engine.Engine()
        e.setup(scenes, None, None, None)
        r = []
        # allow input to be a single event, as well as a sequence of events
        if not misc.issequence(events):
            events = [events]
        # process each event, append each list of output events to the
        # returned list of lists
        for ev in events:
            ret = e.process_event(ev)[:]
            for rev in ret:
                rev.__class__ = MidiEvent
            r.append(ret)
        return r

    def _rebuild_repr(self, scenes):
        """
        Build a string representation of the given scenes using repr(), and
        return new scenes created by evaluating these representations using
        eval().
        """
        r = {}
        for k, v in scenes.items():
            rep = repr(v)
            if 'Process' in rep or 'Call' in rep:
                # patches with Process() units etc. are too tricky for now
                return None
            w = eval(rep, self.mididings_dict)
            # the repr() of the rebuilt patch should be identical to the repr()
            # string it was built from
            self.assertEqual(repr(w), rep)
            r[k] = w
        return r

    def make_event(self, *args, **kwargs):
        """
        Create a new MIDI event. Attributes can be specified in args or
        kwargs, unspecified attributes are filled with random values.
        """
        type, port, channel, data1, data2, sysex = \
            itertools.islice(itertools.chain(args, itertools.repeat(None)), 6)

        for k, v in kwargs.items():
            if k == 'type': type = v
            if k == 'port': port = v
            elif k == 'channel': channel = v
            elif k in ('data1', 'note', 'ctrl'): data1 = v
            elif k in ('data2', 'velocity', 'value'): data2 = v
            elif k == 'program': data2 = v - setup.get_config('data_offset')
            elif k == 'sysex': sysex = v

        if type is None:
            if channel is None:
                type = random.choice(list(set(constants._EVENT_TYPES.values())
                                        - set([DUMMY])))
            else:
                type = random.choice([NOTEON, NOTEOFF, CTRL, PITCHBEND,
                                      AFTERTOUCH, POLY_AFTERTOUCH, PROGRAM])
        elif type == NOTE:
            type = random.choice([NOTEON, NOTEOFF])
        elif misc.issequence(type):
            type = random.choice(type)

        if port is None:
            port = random.randrange(0, 42) + setup.get_config('data_offset')
        if channel is None:
            channel = random.randrange(0, 16) + setup.get_config('data_offset')
        if data1 is None:
            data1 = random.randrange(0, 128)
        if data2 is None:
            data2 = (random.randrange(1, 128) if type == NOTEON
                     else 0 if type == NOTEOFF
                     else random.randrange(0, 128))

        if type == SYSEX and sysex is None:
            sysex = [0xf0] + [random.randrange(0, 128)
                              for n in range(random.randrange(1024))] + [0xf7]

        return MidiEvent(type, port, channel, data1, data2, sysex)

    def modify_event(self, ev, **kwargs):
        """
        Make a copy of the event ev, replacing arbitrary attributes with the
        values given in kwargs.
        """
        r = copy.copy(ev)
        for k, v in kwargs.items():
            setattr(r, k, v)
        return r

    def native_sysex(self, sysex):
        if isinstance(sysex, str):
            sysex = map(ord, sysex)

        return bytearray(sysex)
